/*
 * OggPacket will carry all relevant information of an ogg packet
 *
 * Copyright (C) 2008 Joern Seger
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <iostream>
#include <sstream>
#include <cstring>

#include "oggPacket.h"
#include "log.h"

OggPacketInternal::OggPacketInternal()
  : oggPacket {/*packet*/0,/*bytes*/0, /*b_o_s*/0, /*e_o_s*/0, /*granualpos*/-1, /*packetno*/0},
streamType(OggType::unknown), streamNo(255), streamHeader(false)
{
}

OggPacketInternal::OggPacketInternal(uint8* data, uint32 length,
                                     uint32 packetNo, int64 granulePos, PacketType packetType)
  : oggPacket {/*packet*/data,/*bytes*/length, /*b_o_s*/0, /*e_o_s*/0, /*granualpos*/granulePos, /*packetno*/packetNo},
streamType(OggType::unknown), streamNo(255), streamHeader(false)
{
  switch (packetType) {
  case PacketType::bos:
    oggPacket.b_o_s = 256;
    break;
  case PacketType::eos:
    oggPacket.e_o_s = 256;
    break;
  default: {
  }
  }

}

OggPacket OggPacketInternal::clone()
{
  /* a bit nasty, as ogg_packet is c */
  uint8* data = new uint8[oggPacket.bytes];
  memcpy(data, oggPacket.packet, oggPacket.bytes);

  PacketType packetType(PacketType::normal);

  if (oggPacket.b_o_s)
    packetType = PacketType::bos;

  if (oggPacket.e_o_s)
    packetType = PacketType::eos;

  OggPacket pkt = std::make_shared<OggPacketInternal>(data, oggPacket.bytes, oggPacket.packetno, oggPacket.granulepos, packetType);

  pkt->streamNo = streamNo;
  pkt->streamType = streamType;
  pkt->streamHeader = streamHeader;

  return pkt;
}

OggPacket OggPacketInternal::getPtr()
{
  return shared_from_this();
}

OggPacket OggPacketInternal::create(uint8 *data, uint32 length, uint32 packetNo, int64 granulePos, OggPacketInternal::PacketType packetType)
{
  return std::make_shared<OggPacketInternal>(data, length, packetNo, granulePos, packetType);

}


OggPacketInternal::~OggPacketInternal()
{
  delete[] oggPacket.packet;
}


int64 OggPacketInternal::granulepos()
{
  return oggPacket.granulepos;
}

void OggPacketInternal::setGranulepos(int64 pos)
{
  oggPacket.granulepos = pos;
}

void OggPacketInternal::setStreamHeader()
{
  streamHeader = true;
}

bool OggPacketInternal::isStreamHeader()
{
  return streamHeader;
}

uint32 OggPacketInternal::getPacketNo()
{
  return oggPacket.packetno;
}

uint32 OggPacketInternal::length()
{
  return oggPacket.bytes;
}

bool OggPacketInternal::isBOS()
{
  return oggPacket.b_o_s;
}

bool OggPacketInternal::isEOS()
{
  return oggPacket.e_o_s;
}

void OggPacketInternal::setBOS()
{
  oggPacket.b_o_s = 1;
}

void OggPacketInternal::setEOS()
{
  oggPacket.e_o_s = 1;
}

void OggPacketInternal::unsetBOS()
{
  oggPacket.b_o_s = 0;
}

void OggPacketInternal::unsetEOS()
{
  oggPacket.e_o_s = 0;
}

/*
ogg_packet OggPacketInternal::toLibogg()
{
  return(*objPtr);
}
*/
uint8 OggPacketInternal::getStreamNo()
{
  return streamNo;
}

OggType OggPacketInternal::getStreamType()
{
  return streamType;
}

void OggPacketInternal::setStreamNo(int8 no)
{
  streamNo = no;
}

void OggPacketInternal::setStreamType(OggType type)
{
  streamType = type;
}

/*
void OggPacketInternal::fromLibogg(ogg_packet pack)
{
  // copy all information including the pointers
  bytes      = pack.bytes;
  packetno   = pack.packetno;
  granulepos = pack.granulepos;
  b_o_s      = pack.b_o_s;
  e_o_s      = pack.e_o_s;

  packet = new uint8[pack.bytes];

  if (packet) {
    memcpy(packet, pack.packet, pack.bytes);
  }
  else {
    logger.error() << "OggPacketInternal::fromLibogg: out of memory\n";
  }
}
*/

uint8* OggPacketInternal::data()
{
  return oggPacket.packet;
}

/* print levels:
 * 1) only data length information
 * 2) header information
 * 3) additional header information
 * 4) header dump
 * 5) body dump
 */
std::string OggPacketInternal::print(uint8 level)
{
  std::stringstream retStream;

  retStream << "\nOgg Packet: packet length = " << oggPacket.bytes << std::endl;

  if (level < 1)
    return(retStream.str());

  retStream << "\nHeader Information:"
            << "\n\tBegin of Stream     : ";

  if (oggPacket.b_o_s)
    retStream << "true";
  else
    retStream << "false";

  retStream << "\n\tEnd of Stream       : ";

  if (oggPacket.e_o_s)
    retStream << "true";
  else
    retStream << "false";

  retStream << "\n\tGranule Position    : " << oggPacket.granulepos;
  retStream << "\n\tPacket Number       : " << oggPacket.packetno;

  retStream << std::endl;

  if (level < 3)
    return(retStream.str());

  retStream << "\n\tStream Number       : " << (int)streamNo;
  retStream << "\n\tStream Type         : ";

  switch (streamType) {
  case OggType::vorbis:
    retStream << "Vorbis";
    break;
  case OggType::theora:
    retStream << "Theora";
    break;
  case OggType::kate:
    retStream << "Kate";
    break;
  case OggType::unknown:
  default:
    retStream << "unknown";
    break;
  }
  retStream << std::endl;

  if (level < 4)
    return(retStream.str());

  retStream << "\nPacket Hex dump:" << std::hex;

  for (int32 c(0); c<oggPacket.bytes; ++c) {
    if ((c%16) == 0)
      retStream << std::endl;
    if (((unsigned) (oggPacket.packet[c])) < 16)
      retStream << "0";
    retStream << (unsigned) (oggPacket.packet[c]);
  }

  retStream << std::dec << std::endl;

  return(retStream.str());
}
